<?php
/*

  While developing this software, I discovered a bug which appears in
  PHP 5 on Debian Linux (to be precise, in version 5.2.6.dfsg.1-1+lenny8
  of the package php5-cli), but which doesn't appear in PHP 5.2.13.
  The bug in question is probably this one:

    http://bugs.php.net/bug.php?id=45181

  I worked around it by calling clearstatcache after every call to chdir.
  For safety's sake, I then ensured that every call to `stat' is preceded
  by a call to clearstatcache.

*/

# these constants are for internal use only 
#
define("DIR_TREE_WALK_CHDIR_SUCCESSFUL", 1);
define("DIR_TREE_WALK_POSSIBLE_ATTACK", 2);
define("DIR_TREE_WALK_INACCESSIBLE", 3);

function dir_tree_walk($top, $process_func, $options = array()) {
  if(isset($options["prune"]) && $options["prune"] &&
    !array_key_exists("preprocess", $options)) {

    trigger_error("When calling " . __FUNCTION__ .
      ", it makes no sense to " .
      "set `prune' to TRUE without also setting `preprocess'",
      E_USER_WARNING);
  }

  $defaults = array(
    "prune" => FALSE,
    "preprocess" => NULL,
    "postprocess" => NULL,
    "chdir_failure_action" => NULL,
    "top_down" => TRUE,
    "follow_links" => FALSE
  );

  extract(override_options($defaults, $options));

  $top = make_path_safe($top);

  if(!file_exists($top)) {
    throw new InvalidTopException("$top doesn't exist");
  } elseif(!is_dir($top)) {
    throw new InvalidTopException("$top is not a directory");
  } else {
    $cur_dir = getcwd_safe();

    $dir_saver = new dir_saver($cur_dir);

    $top_abs = $top{0} == "/"? $top: "$cur_dir/$top";

    if($top_down)
      call_user_func($process_func, $top, $top, $top_abs);

    try {
      dir_tree_walk_aux($top, $top, $top_abs, $cur_dir, $process_func,
	$prune, $preprocess, $postprocess, $chdir_failure_action, $top_down,
	$follow_links);
    } catch (Exception $e) {
      $dir_saver->go_back();
      throw $e;
    }

    if(!$top_down)
      call_user_func($process_func, $top, $top, $top_abs);
  }
}

function dir_tree_walk_aux($new_dir1, $new_dir2, $new_dir_abs, $cur_dir,
  $process_func, $prune, $preprocess, $postprocess, $chdir_failure_action,
  $top_down, $follow_links) {

  /*
    we call is_link here.  it is possible for an attacker to
    wait until is_link has been called, then turn $e into a symlink.
    this attack is handled in chdir_and_check.
  */
  if($follow_links || !is_link($new_dir1)) {
    $dir_saver = new dir_saver($cur_dir);

    if(chdir_securely($new_dir1, $new_dir2, $new_dir_abs,
      $chdir_failure_action, $follow_links)) {

      $entries = array();
      $handle = opendir_exc(".");
      while(TRUE) {
	$e = readdir($handle);
	if($e === FALSE) break;
	if($e != "." && $e != "..")
	  array_push($entries, $e);
      }
      closedir($handle);

      if(isset($preprocess))
	$entries = call_user_func($preprocess, $entries, $new_dir2,
	  $new_dir_abs);

      foreach ($entries as $e) {
	$is_dir = is_dir($e);
	$process_before_entering = $top_down || !$is_dir;
	$prune_this_entry = FALSE;
	if($process_before_entering) {
	  $temp = (boolean) call_user_func($process_func, $e, "$new_dir2/$e",
	    "$new_dir_abs/$e");
	  $prune_this_entry = $prune && $temp;
	}
	if($is_dir && !$prune_this_entry) {
	  dir_tree_walk_aux($e, "$new_dir2/$e", "$new_dir_abs/$e",
	    $new_dir_abs, $process_func, $prune, $preprocess,
	    $postprocess, $chdir_failure_action, $top_down,
	    $follow_links);
	}
	if(!$process_before_entering)
	  call_user_func($process_func, $e, "$new_dir2/$e",
	    "$new_dir_abs/$e");
      }

      if(isset($postprocess))
	call_user_func($postprocess, $new_dir2, $new_dir_abs);

      $dir_saver->go_back();
    }
  }
}

class dir_saver {
  function __construct($dir) {
    $this->saved_stat_info = stat_no_cache($dir);
    $this->dir = $dir;
  }

  function go_back() {
    if(return_to_dir($this->dir)) {
      if(!same_filesystem_entity(stat_no_cache("."),
	$this->saved_stat_info)) {

	throw new DirSwitchException("cannot securely return to " .
	  "directory $this->dir because it appears to be a " .
	  "different directory (but with the same name)");
      }
    }
  }
}

function same_filesystem_entity($stat_info1, $stat_info2) {
  return is_array($stat_info1) && is_array($stat_info2) &&
    $stat_info1["dev"] === $stat_info2["dev"] &&
    $stat_info1["ino"] === $stat_info2["ino"];
}

/*
  the idea for the following function came from the GNU `find'
  documentation:

    http://www.gnu.org/software/findutils/manual/html_node/find_html/Systems-without-O_005fNOFOLLOW.html#Systems-without-O_005fNOFOLLOW
*/
function chdir_securely($dir, $dir2, $dir_abs, $chdir_failure_action,
  $follow_links) {

  for($attempt = 1; ; $attempt++)
    switch(chdir_and_check($dir, $dir2, $dir_abs, $chdir_failure_action,
      $follow_links)) {

      case DIR_TREE_WALK_CHDIR_SUCCESSFUL:
	return TRUE;
      ;;
      case DIR_TREE_WALK_INACCESSIBLE:
	return FALSE;
      ;;
      case DIR_TREE_WALK_POSSIBLE_ATTACK:
	if($attempt == 2) {
	  trigger_error("apparent attempt to attack program",
	    E_USER_WARNING);
	  return FALSE;
	}
      ;;
    }
}

function chdir_and_check($dir, $dir2, $dir_abs, $chdir_failure_action,
  $follow_links) {

  $stat_func = $follow_links? "stat_no_cache": "lstat_no_cache";

  $dir_saver = new dir_saver($dir_abs);
  $stat_info = $stat_func($dir);

  if(!chdir2($dir)) {
    $err_msg = isset($php_errormsg)? $php_errormsg: NULL;

    if(!is_null($chdir_failure_action))
      call_user_func($chdir_failure_action, $dir2, $dir_abs, $err_msg);

    return DIR_TREE_WALK_INACCESSIBLE;
  } else {
    if(same_filesystem_entity($stat_func("."), $stat_info)) {
      return DIR_TREE_WALK_CHDIR_SUCCESSFUL;
    } else {
      $dir_saver->go_back();
      return DIR_TREE_WALK_POSSIBLE_ATTACK;
    }
  } 
}

class OSException extends RuntimeException {}
class ReturnToDirException extends OSException {}
class OpendirException extends OSException {}

class SecurityException extends RuntimeException {}

class DirSwitchException extends SecurityException {}

class InvalidTopException extends RuntimeException {}

function return_to_dir($dir) {
  $t = error_to_exception("chdir", $dir, "ReturnToDirException");
  clearstatcache();
  return $t;
}

function opendir_exc($dir) {
  return error_to_exception("opendir", $dir, "OpendirException");
}

function error_to_exception($func, $arg, $exception_class) {
  $result = $func($arg);

  if($result === FALSE) {
    $msg = isset($php_errormsg)? $php_errormsg: "$func failed on $arg";
    throw new $exception_class($msg);
  } else
    return $result;
}

function override_options($default, $specified) {
  $new = $default;

  foreach ($specified as $key => $val)
    if(array_key_exists($key, $default))
      $new[$key] = $val;

  return $new;
}

/*

  make_path_safe: this function changes a path to ensure that no PHP
  "stream wrappers" will be invoked.

  Consider a path such as "http://a/b/c"; this will obviously be
  considered by PHP as a HTTP URL.  Or consider a directory called
  "http:", which has been specified as "http:/" and which contains a
  subdirectory called www.example.com; it is possible that some code could
  join "http:/" and "www.example.com" to form "http://www.example.com".

  Is the logic used here sufficient to eliminate the problem?  Yes,
  in view of the following sentence from the PHP documentation:

    The URL syntax used to describe a wrapper only supports the
    scheme://... syntax. scheme:/ and scheme: syntaxes are not supported.

*/
function make_path_safe($path) {
  $path = preg_replace('@//+@', "/", $path);
  return $path === "/"? "/": rtrim($path, "/");
}

# this function is probably unnecessary, but it won't hurt
#
function getcwd_safe() {
  return make_path_safe(getcwd());
}

function stat_no_cache($filename) {
  clearstatcache();
  return stat($filename);
}

function lstat_no_cache($filename) {
  clearstatcache();
  return lstat($filename);
}

function chdir2($dir) {
  $t = chdir($dir);
  clearstatcache();
  return $t;
}
