<?php
/*
  Here are 6 fairly brief examples of how to use
  dir_tree_walk.  Many of them would be shorter
  and prettier if we could use closures, but closures
  seem to be buggy in PHP 5.3.2 (the latest version
  at time of writing).
*/

require_once "dir_tree_walk.php";

$document_root = "/var/www";

/*
  Example 1: get the total size of all files under the document root,
  excluding JPEG files.  Print a running total as each directory
  is completed.
*/

$size_summer = new size_summer;

dir_tree_walk($document_root,
  array($size_summer, "add_size"),
  array(
    "preprocess" => "exclude_jpeg",
    "postprocess" => array($size_summer, "print_total_so_far")));

print "Total size of non-JPEG files: $size_summer->size\n";

class size_summer {
  var $size = 0;

  function add_size($entity) {
    if(is_file($entity))
      $this->size += filesize($entity);
  }

  function print_total_so_far($dir) {
    printf("Directory: %s; total so far: %s\n", $dir, $this->size);
  }
}

function exclude_jpeg($list) {
  # we are assuming that no directory has an extension such as ".jpg"
  #
  return preg_grep('@.\.(jpg|jpeg)\z@i', $list, PREG_GREP_INVERT);
}

/*
  Example 2: get a list of all the PHP files under the document root.
*/

$PHP_lister = new PHP_lister;

dir_tree_walk($document_root, array($PHP_lister, "add_PHP_file"));

print "These are the PHP files:\n" .
  implode("\n", $PHP_lister->list) . "\n";

class PHP_lister {
  var $list = array();

  function add_PHP_file($entity) {
    if(preg_match('@.\.php\z@', $entity))
      array_push($this->list, $entity);
  }
}

/*
  Example 3: delete every file under /var/www that looks like a temporary
  file.
*/

dir_tree_walk($document_root, "delete_tmp_file");

function delete_tmp_file($entity) {
  if(preg_match('@^tmp.@', $entity) && is_file($entity))
    unlink($entity);
}

/*
  Example 4: find out the number of files under a Subversion "working
  copy", not including Subversion's own data (i.e. not including
  the ".svn" directories and everything under them).
*/

$ignorer = new svn_ignorer;

dir_tree_walk($_ENV["HOME"] . "/my-code", array($ignorer, "examine"),
  array("prune", TRUE));
  /*
    setting "prune" to TRUE tells dir_tree_walk not to descend
    into any directory for which `examine' returned a true value.
    (If "prune" were not set to TRUE, then the return value of
    `examine' would simply be ignored.)
  */

print "Number of files under working copy: $ignorer->count\n";

class svn_ignorer {
  var $count = 0;

  function examine($entity) {
    if(preg_match('@(^|/)\.svn\z@', $entity) && is_dir($entity)) {
      return TRUE;
    } else {
      $this->count++;
      return FALSE;
    }
  }
}

/*
  Example 5: attempt to find a file under the document root which
  contains the string "Jupiter"; if one is found, don't examine
  any more files.
*/

$finder = new string_finder("Jupiter");

dir_tree_walk($document_root, array($finder, "look"),
  array("preprocess" => array($finder, "stop_after_success")));

if($finder->file_found()) {
  print "Found file: " . $finder->file() . "\n";
} else
  print "No file found\n";

class string_finder {
  function __construct($string) {
    $this->string = $string;
    $this->found = FALSE;
  }

  function look($entity) {
    if(!$this->found && is_file($entity)) {
      if(is_integer(strpos(file_get_contents($entity), $this->string))) {
        $this->found = TRUE;
        $this->file = $entity;
      }
    }
  }

  function stop_after_success($entries) {
    return $this->found? array(): $entries;
  }

  function file() {
    return $this->file;
  }

  function file_found() {
    return $this->found;
  }
}

/*
  The above example may do a good deal of unnecessary work.  This is
  especially likely if directories tend to have a large number of files
  and if we have to descend a long way to find a suitable file.

  The following may be faster, though it is a disgusting abuse of
  the exception mechanism.
*/

/*
  Example 6: does the same as Example 5, but hopefully faster.
*/

$finder2 = new string_finder2("Jupiter");

unset($found_file);

try {
  dir_tree_walk($document_root, array($finder2, "look"));
} catch(StringFoundException $e) {
  $found_file = $e->getMessage();
}

if(isset($found_file)) {
  print "Found file: $found_file\n";
} else {
  print "No file found\n";
}

class string_finder2 {
  function __construct($string) {
    $this->string = $string;
  }

  function look($entity) {
    if(is_file($entity) &&
      is_integer(strpos(file_get_contents($entity), $this->string))) {

      throw new StringFoundException($entity);
    }
  }
}

class StringFoundException extends Exception {}
